import nltk


def download_nltk_data(quiet=False):
    downloads = (
        "stopwords",
        "punkt",
        "wordnet",
        "averaged_perceptron_tagger",
        "omw-1.4",
    )

    for d in downloads:
        nltk.download(d, quiet=quiet)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ttc', 'ttc.chatbot', 'ttc.cli']

package_data = \
{'': ['*']}

install_requires = \
['contractions>=0.1.72,<0.2.0',
 'inquirerpy>=0.3.4,<0.4.0',
 'nltk>=3.7,<4.0',
 'typer>=0.6.1,<0.7.0',
 'yaspin>=2.2.0,<3.0.0']

entry_points = \
{'console_scripts': ['ttc = ttc.cli.main:app']}

setup_kwargs = {
    'name': 'thomasthechatbot',
    'version': '1.0.2',
    'description': 'A Python chatbot that learns as you speak to it.',
    'long_description': '<div align="center">\n    <img src="https://i.imgur.com/hA9YF2s.png" alt="Thomas" width="220" height="220">\n    <h1>Thomas the Chatbot</h1>\n</div>\n\n![Demo](https://i.imgur.com/Jet4UGh.gif)\n\n# Installation\n\n**Python 3.9+ is required**\n\nThis package can be installed from [PyPi](https://pypi.org/project/thomasthechatbot/) with:\n\n```\npip install thomasthechatbot\n```\n\n# Usage\n\n## Basic Usage\n\n```py\nfrom ttc import Chatbot, Context, download_nltk_data\n\n# Only needs to be run once (can be removed after first run)\ndownload_nltk_data()\n\n# Creating the context\nctx = Context()\n\n# Initializing the chatbot\nchatbot = Chatbot()\n\ntalk = True\n\nwhile talk:\n    msg = input("You: ")\n\n    if msg == "s":\n        talk = False\n    else:\n        # Getting the response\n        resp = chatbot.respond(ctx, msg)\n\n        # Saving the response to the context\n        ctx.save_resp(resp)\n\n        print(f"Thomas: {resp}")\n\n# Saving the chatbot data\nchatbot.save_data()\n```\n\n## Configurations\n\n```py\nchatbot = Chatbot(\n    path="brain",\n    learn=False,\n    min_score=0.5,\n    score_threshold=0.5,\n    mesh_association=0.5,\n)\n```\n\n## CLI\n\nType `ttc` to begin talking to Thomas.\n\n# How does Thomas work?\n\nThomas has no hard-coded responses and is designed to “learn” as he is spoken to.\n\nNote: I created this approach based on my intuition and not a proven method.\n\n## Data Storage\n\nThomas does not come up with his own responses, he reiterates those that he has seen before.\n\n### Responses\n\nPrevious responses are stored in `resps.json` as a dictionary where the key is a generated [UUID](https://docs.python.org/3/library/uuid.html) and the value is the tokenized response.\n\n### Mesh\n\nPrompts are associated with responses through a "mesh" which is stored in `mesh.thomas`. The mesh consists of a dictionary where the key is the UUID of the prompt and the value is a "link". Links associate responses to patterns of words, they have the following attributes:\n\n`stop_words: set`\nStop words separated from the tokenized prompt.\n\n`keywords: set`\nThe remaining words which are lemmatized by their part of speech.\n\n`resps: dict[str, set]`\nResponses to the prompt where the key is the response UUID and the value is a set of mesh ids from the previous prompt.\n\n## Querying Responses\n\n### Tokenizing Prompts\n\nBefore tokenization, prompts are lowercased, contractions are expanded and punctuation is removed. This aids in improving the consistency and accuracy of queries. Prompts are tokenized by word and split into key words and stop words.\n\n### Ignoring Responses\n\nThe user\'s prompt and chatbot\'s previous response are ignored to prevent the chatbot from appearing repetitive.\n\n### Initial Query\n\nMeshes are initially queried by their score which can be calculated with:\n\n`(ss / 2 + sk) / (ts / 2 + tk - ss / 2 - sk + 1)`\n\n`ss` = shared stop words\n\n`sk` = shared key words\n\n`ts` = total stop words\n\n`tk` = total key words\n\nThis formula weighs shared key words 2 times more heavily than stop words by dividing `ss` and `sk` by 2. It also takes into account the total number of words resulting in more precise meshes being favoured.\n\n### First Discard\n\nMeshes with scores below a threshold (`min_score`) are discarded.\n\n### No Results Queried\n\nIf no results remain, meshes are queried by the number of shared stop words.\n\n### Second Discard\n\nThe remaining meshes are sorted and meshes that fall below a percentage threshold (`score_threshold`) of the best score are discarded. Considering multiple meshes increases the variety of responses.\n\n### Mesh Association\n\nMeshes are associated with each other by the percentage of shared responses (`mesh_association`). Associated meshes for each queried mesh are found and added to the list. This process prevents less trained prompts from having a small response pool.\n\n### Choosing a Response\n\nIf responses are found to share the same previous message UUID as the prompt, all non-sharing responses are moved. Responses are chosen at random from the remaining responses. Random selection prevents the chatbot from being predictable.\n\n# Contributing\n\nOpen to contributions, please create an issue if you want to do so.\n\n# Formatting\n\n[Black](https://github.com/psf/black), [isort](https://github.com/PyCQA/isort) and [Prettier](https://prettier.io/) are used for formatting\n',
    'author': 'principle105',
    'author_email': 'principle105@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/principle105/thomasthechatbot',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""
MIT License

Copyright (c) 2022 Principle

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import json
import os
import random
import string
import uuid
from collections import defaultdict
from typing import TYPE_CHECKING, Callable

import contractions
from nltk import pos_tag, word_tokenize

if TYPE_CHECKING:
    from .context import Context

# Constants
DEFAULT_STORAGE_PATH = "storage"


def _load_json_file(path: str, name: str) -> dict:
    directory = f"{path}/{name}.json"

    if not os.path.exists(directory):
        return {}

    with open(directory, "r") as f:
        return json.load(f)


def _save_json_file(path: str, name: str, data: dict):
    directory = f"{path}/{name}.json"

    if not os.path.exists(path):
        os.makedirs(path)

    with open(directory, "w") as f:
        return json.dump(data, f)


def _generate_uuid():
    return str(uuid.uuid1())


def _dict_values_to_set(d: list[dict]) -> set:
    """Converts the values of a list of dictionaries to a set."""
    return set().union(*d.values())


class Storage:
    file_name: str = ...

    def __init__(self, data: dict):
        self.data = data

    def to_json(self) -> dict:
        ...

    @classmethod
    def from_json(cls, data: dict):
        ...

    def save_data(self, path: str):
        json_data = self.to_json()

        _save_json_file(path, self.file_name, json_data)

    @classmethod
    def from_file(cls, path: str):
        json_data = _load_json_file(path, cls.file_name)

        return cls.from_json(json_data)


class Link:
    def __init__(self, keywords: set, stop_words: set, resps: dict[str, set] = None):
        if resps is None:
            resps = {}

        # Query data
        self.keywords = keywords
        self.stop_words = stop_words

        # Responses {response_uuid: {previous_uuid}}
        self.resps = resps

    def to_json(self) -> dict:
        return {
            "keywords": list(self.keywords),
            "stop_words": list(self.stop_words),
            "resps": {v: list(k) for v, k in self.resps.items()},
        }

    @classmethod
    def from_json(cls, data: dict):
        return cls(
            keywords=set(data["keywords"]),
            stop_words=set(data["stop_words"]),
            resps={v: set(k) for v, k in data["resps"].items()},
        )


class Mesh(Storage):
    file_name = "mesh"

    def __init__(self, data: dict[str, Link]):
        super().__init__(data)

    def to_json(self) -> dict:
        return {k: v.to_json() for k, v in self.data.items()}

    @classmethod
    def from_json(cls, data: dict):
        return cls({k: Link.from_json(v) for k, v in data.items()})

    def find_mesh(
        self, score_callback: Callable[[Link], int], min_score: int, ignore: set = None
    ):

        if ignore is None:
            ignore = set()

        for _id, data in self.data.items():

            score = score_callback(data)

            if score < min_score:
                continue

            resps_left = set(data.resps) - ignore

            if not resps_left:
                continue

            yield _id, data, resps_left, score

    def add_new_mesh(self, _id: str, keywords: set, stop_words: set):
        if _id is None:
            _id = _generate_uuid()

        link = Link(keywords, stop_words)

        self.data[_id] = link

    def add_mesh_resp(self, _id: str, resp: str, prev_resp: str = None):
        if resp not in self.data[_id].resps:
            self.data[_id].resps[resp] = set()

        if prev_resp is not None:
            self.data[_id].resps[resp].add(prev_resp)

    def resp_id_in_mesh(self, _id: str) -> bool:
        return _id in self.data

    def resp_id_in_mesh_resp(self, _id: str, resp: str) -> bool:
        return resp in self.data[_id].resps

    def get_mesh_from_id(self, _id: str) -> Link:
        return self.data.get(_id, None)


class Resps(Storage):
    file_name = "resps"

    def __init__(self, data: dict[str, list]):
        super().__init__(data)

    def to_json(self) -> dict:
        return self.data

    @classmethod
    def from_json(cls, data: dict):
        return cls(data)

    def add_new_resp(self, msg_tokens: list) -> str:
        _id = _generate_uuid()

        self.data[_id] = msg_tokens

        return _id

    def get_resp_from_tokens(self, msg_tokens: list) -> str:
        resp_values = tuple(self.data.values())

        if msg_tokens not in resp_values:
            return None

        resp_index = resp_values.index(msg_tokens)

        return tuple(self.data.keys())[resp_index]

    def get_resp_from_id(self, _id: str):
        return self.data.get(_id, None)


class Response:
    def __init__(self, resp: str, resp_id: str, mesh_id: str = None):
        self.resp = resp
        self.resp_id = resp_id
        self.mesh_id = mesh_id

    def __str__(self) -> str:
        return self.resp


class Chatbot:
    def __init__(
        self,
        *,
        path: str = None,
        learn: bool = True,
        min_score: float = 0.7,
        score_threshold: float = 0.7,
        mesh_association: float = 0.6,
    ):

        from nltk.corpus import stopwords
        from nltk.stem.wordnet import WordNetLemmatizer

        # The path to the storage directory
        self.path = path or DEFAULT_STORAGE_PATH

        # The states conrolling thomas' memory
        self.mesh = Mesh.from_file(self.path)
        self.resps = Resps.from_file(self.path)

        # Configurations
        self.learn = learn
        self.min_score = min_score
        self.score_threshold = score_threshold
        self.mesh_association = mesh_association

        # Responsible for lemmatizing words
        self.tag_map = self.create_tag_map()
        self.wl = WordNetLemmatizer()

        self.all_stop_words = set(stopwords.words("english"))

    def create_tag_map(self):
        from nltk.corpus import wordnet

        tag_map = defaultdict(lambda: wordnet.NOUN)
        tag_map["J"] = wordnet.ADJ
        tag_map["V"] = wordnet.VERB
        tag_map["R"] = wordnet.ADV

        return tag_map

    def tokenize_msg(self, msg: str) -> tuple[set]:
        text = msg.lower()

        # Removing contrations
        text = contractions.fix(text)

        # Removing punctuation
        text = text.translate(str.maketrans("", "", string.punctuation))

        return word_tokenize(text)

    def lemmatize_tokens(self, tokens: list):
        for token, tag in pos_tag(tokens):
            yield self.wl.lemmatize(token, self.tag_map[tag[0]])

    def separate_tokens(self, raw_tokens: list):
        stop_words = set()
        tokens = set()

        # Separating stop words
        for w in self.lemmatize_tokens(raw_tokens):
            if w in self.all_stop_words:
                stop_words.add(w)
            else:
                # Stemming helps with matching key words
                tokens.add(w)

        return tokens, stop_words

    def _parse_resps(self, resps, ignore: set = None) -> dict[str, list]:
        if ignore is None:
            ignore = set()

        return {d: self.resps.data[d] for d in resps if d not in ignore}

    def get_unlinked_resps(self, ignore: set = None):
        """Finds responses that do not have a response in the mesh"""
        if ignore is None:
            ignore = set()

        resp_ids = set(self.resps.data)
        mesh_ids = set(self.mesh.data)

        no_mesh_resp = resp_ids - mesh_ids

        return self._parse_resps(no_mesh_resp, ignore)

    def get_all_resps(self, ignore: set = None):
        return self._parse_resps(self.resps.data, ignore)

    def tokenize(self, text: str):
        # Tokenizing the input
        raw_tokens = self.tokenize_msg(text)

        # Separating the stop words from the key words
        keywords, stop_words = self.separate_tokens(raw_tokens)

        return raw_tokens, keywords, stop_words

    def find_elligible_meshes(self, meshes: list, threshold: float):
        # Sorting the results by the check function
        sorted_results = sorted(meshes, key=lambda x: x[3], reverse=True)

        best_score = sorted_results[0][3]
        min_score = best_score * threshold

        # Picking the top responses that are within a percentage threshold of the best one
        return [r for r in sorted_results if r[3] >= min_score]

    def find_resps_from_last_msg(
        self, resps_left: set, resps: dict, last_msg: str
    ) -> set:
        return {_id for _id in resps_left if last_msg in resps[_id]}

    def respond(self, ctx: "Context", msg: str) -> Response:
        raw_tokens, keywords, stop_words = self.tokenize(msg)

        msg_id = self.resps.get_resp_from_tokens(raw_tokens)

        if self.learn:
            # Saving the user's message
            if msg_id is None:
                msg_id = self.resps.add_new_resp(msg_tokens=raw_tokens)

            save_resp = True

            # Saving the user's mesasge as if it was a response to the bot
            if ctx.last_resp is not None:
                if not self.mesh.resp_id_in_mesh(ctx.last_resp):
                    resp = self.resps.get_resp_from_id(ctx.last_resp)

                    k, s = self.separate_tokens(resp)

                    self.mesh.add_new_mesh(ctx.last_resp, k, s)

                # Checking if the response is already in the mesh
                elif self.mesh.resp_id_in_mesh_resp(ctx.last_resp, msg_id):
                    save_resp = False

                if save_resp:
                    self.mesh.add_mesh_resp(ctx.last_resp, msg_id, ctx.last_msg)

        ignore = set()

        # Prevents the bot from responding with the same message twice
        if ctx.last_resp is not None:
            ignore.add(ctx.last_resp)

        # Prevents the bot from saying the same thing as the user
        if msg_id is not None:
            ignore.add(msg_id)

        def _filter(r: Link):
            t = len(stop_words & r.stop_words) / 2 + len(r.keywords & keywords)

            return t / (len(r.stop_words) / 2 + len(r.keywords) - t + 1)

        results = [
            k
            for k in self.mesh.find_mesh(
                _filter, min_score=self.min_score, ignore=ignore
            )
        ]

        if results == []:
            # Trying the query just using stop words
            _filter = lambda r: len(stop_words & r.stop_words)

            results = [
                k for k in self.mesh.find_mesh(_filter, min_score=1, ignore=ignore)
            ]

            # Picking a random response if there are still no results
            if results == []:
                prev_meshes = self.get_unlinked_resps(ignore=ignore)

                if not prev_meshes:
                    # Getting all the responses if there are no unlinked responses

                    resp_ignore = ignore if len(ignore) < len(self.resps.data) else None

                    prev_meshes = self.get_all_resps(ignore=resp_ignore)

                resp_ids = tuple(prev_meshes.keys())

                if not resp_ids:
                    raise Exception("No responses found, configure learning to True")

                # Picking a random response
                resp_id = random.choice(resp_ids)

                resp = " ".join(prev_meshes[resp_id])

                return Response(resp=resp, resp_id=resp_id)

        # Finding elligible meshes from keywords
        results = self.find_elligible_meshes(results, self.score_threshold)

        # Mesh responses from previous message
        prev_meshes = {}

        # All mesh responses
        all_meshes = {}

        # Trying to find a response by the user's previous message
        for mesh_id, link, resps_left, _ in results:
            r = set()

            if ctx.last_msg is not None:
                # Finding the response ids that have the same previous message id
                r = self.find_resps_from_last_msg(resps_left, link.resps, ctx.last_msg)

                if r:
                    prev_meshes[mesh_id] = prev_meshes.get(mesh_id, set()) | r

            all_meshes[mesh_id] = all_meshes.get(mesh_id, set()) | resps_left

        meshes = prev_meshes if prev_meshes else all_meshes

        initial_resps = _dict_values_to_set(meshes)
        share_threshold = len(initial_resps) * self.mesh_association

        _filter = lambda x: len(set(x.resps) & initial_resps)

        # Finding other meshes that share a percentage of similar responses
        for mesh_id, link, resps_left, _ in self.mesh.find_mesh(
            _filter, min_score=share_threshold, ignore=ignore
        ):
            resps = meshes.get(mesh_id, set())

            if ctx.last_msg is not None:
                r = self.find_resps_from_last_msg(resps_left, link.resps, ctx.last_msg)

                if r:
                    resps |= r

            if not prev_meshes:
                resps |= set(resps_left)

            meshes[mesh_id] = resps

        # Converting to list of responses to allow for random selection
        resps = _dict_values_to_set(meshes)

        # Choosing a random mesh from the elligible mesh responses
        resp_id = random.choice(tuple(resps))

        # Getting the mesh id
        mesh_id = next((_id for _id, r in meshes.items() if resp_id in r), None)

        resp = self.resps.get_resp_from_id(resp_id)

        resp = " ".join(resp)

        return Response(resp=resp, resp_id=resp_id, mesh_id=mesh_id)

    def save_data(self):
        self.mesh.save_data(self.path)
        self.resps.save_data(self.path)



"""
MIT License

Copyright (c) 2022 Principle

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE."""

import typer
from InquirerPy import inquirer
from InquirerPy.utils import color_print
from InquirerPy.validator import EmptyInputValidator
from yaspin import yaspin

from ttc.utils import download_nltk_data

# Initializing the CLI
app = typer.Typer()


class Sender:
    @staticmethod
    def success(text: str):
        color_print([("green", text)])

    @staticmethod
    def fail(text: str):
        color_print([("red", text)])

    @staticmethod
    def thomas(text: str):
        color_print([("#f6ca44", text)])

    @staticmethod
    def regular(text: str):
        color_print([("", text)])


@app.command()
def start():
    with yaspin(text="Loading", color="cyan") as sp:
        download_nltk_data(quiet=True)
        sp.write("- Downloaded data")

        # Importing chatbot after downloading nltk data to prevent error
        from ttc import Chatbot, Context

        chatbot = Chatbot()
        sp.write("- Brain loaded")

        sp.ok("✔")

    Sender.success(
        "\nThomas learns as you speak to him.\nSave his brain by typing 's'.\n"
    )

    # Creating the user's context
    ctx = Context()

    while True:
        text = inquirer.text(
            message="You:",
            validate=EmptyInputValidator(),
            amark="",
        ).execute()

        if text == "s":
            save = inquirer.confirm(
                message="Are you sure you want to save the new brain?", default=False
            ).execute()

            if save:
                chatbot.save_data()

                Sender.success("Saving data...")
                return

            continue

        with yaspin(color="yellow") as spinner:
            resp = chatbot.respond(ctx, text)

            spinner.stop()

            ctx.save_resp(resp)

            # Sending the response
            Sender.thomas(f"Thomas: {resp}")
